# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Device detector module."""
import copy
import os
import typing
from typing import Any, Dict, List, Optional, Tuple
import weakref

from gazoo_device import config
from gazoo_device import custom_types
from gazoo_device import detect_criteria
from gazoo_device import extensions
from gazoo_device import gdm_logger
from gazoo_device.base_classes import auxiliary_device_base
from gazoo_device.switchboard import communication_types
from gazoo_device.utility import pty_process_utils

WIKI_URL = (
    "https://github.com/google/gazoo-device/blob/master/docs/device_setup")
logger = gdm_logger.get_logger()


class DeviceDetector(object):
  """Class for detecting devices.

  Class scans systems for all connections, filters out known connections,
  then attempts to assign connection to right device classes and create
  entry in configs for persistent info (such as ID and serial_number)
  """

  def __init__(
      self,
      manager: Any,
      log_directory: str,
      persistent_configs: custom_types.PersistentConfigsDict,
      options_configs: custom_types.OptionalConfigsDict,
      supported_auxiliary_device_classes: List[
          auxiliary_device_base.AuxiliaryDeviceBase]):
    """Initializes the device detector.

    Args:
        manager: instance of Manager.
        log_directory: location of logs.
        persistent_configs: devices known to the manager.
        options_configs: device options known to the manager.
        supported_auxiliary_device_classes: list of auxiliary device
            classes.
    """
    self.manager_weakref = weakref.ref(manager)
    self.log_directory = log_directory
    self.auxiliary_classes = supported_auxiliary_device_classes
    self.persistent_configs = copy.deepcopy(persistent_configs)
    self.options_configs = copy.deepcopy(options_configs)
    self.known_connections = self._create_known_connections()

  def detect_all_new_devices(
      self, static_ips: Optional[List[str]] = None
    ) -> Tuple[
        custom_types.PersistentConfigsDict,
        custom_types.OptionalConfigsDict]:
    """Finds all possible new connections and detects devices.

    Args:
        static_ips: Static ips not otherwise detectable.

    Returns:
        (Dicts of persistent props, dict of optional props).
    """
    logger.info(
        "\n##### Step 1/3: Detecting potential new communication addresses. #####\n"
    )
    all_connections_dict = communication_types.detect_connections(static_ips)
    return self.detect_new_devices(all_connections_dict)

  def detect_new_devices(
      self, connections_dict: Dict[str, List[str]]
    ) -> Tuple[
        custom_types.PersistentConfigsDict,
        custom_types.OptionalConfigsDict]:
    """Detects the devices for given connections.

    --> Filters out known connections
    --> Determine device type
    --> Detects persistent info
    --> Updates persistent and optional configs
    --> Tracks errors and unidentified connections and Writes a summary

    Args:
        connections_dict: connection device paths grouped by connection name.
            The connection name corresponds with connection class names
            in communication_types.py

    Returns:
        (persistent_configs, options_configs) for all devices (new and old).
    """

    connections_dict = self._filter_out_known_connections(
        connections_dict, self.known_connections)

    possible_device_tuples, errs, no_id_cons = self._identify_connection_device_class(
        connections_dict)
    new_names = []
    logger.info(
        "\n##### Step 3/3: Extract Persistent Info from Detected Devices. #####\n"
    )
    for device_class, connection in possible_device_tuples:
      try:
        name, persistent_props, optional_props = self._detect_get_info(
            device_class, connection)
        new_names.append(name)
        self._add_to_configs(device_class, name, persistent_props,
                             optional_props)
      except Exception as err:  # pylint: disable=broad-except
        msg = "Error extracting info from {} {!r}. Err: {!r}".format(
            device_class.DEVICE_TYPE, connection, err)
        errs.append(msg)
        no_id_cons.append(connection)

    self._print_summary(new_names, errs, no_id_cons)
    return self.persistent_configs, self.options_configs

  def _add_to_configs(
      self,
      device_class: custom_types.Device,
      name: str,
      persistent_props: custom_types.DeviceConfig,
      optional_props: custom_types.DeviceConfig) -> None:
    """Adds newly detect info to appropriate location in config.

    Args:
        device_class: device class
        name: name of device
        persistent_props: dict of device's persistent props
        optional_props: dict of device's optional props
    """

    if device_class in self.auxiliary_classes:
      ind = 1
      more_props = {key: None for key in config.HUB_OPTION_ATTRIBUTES}
    else:
      ind = 0
      more_props = {key: None for key in config.DEVICE_OPTION_ATTRIBUTES}

    # Add generic props to options
    optional_props.update(more_props)

    self.persistent_configs[config.DEVICES_KEYS[ind]][name] = persistent_props
    self.options_configs[config.OPTIONS_KEYS[ind]][name] = optional_props

  def _create_known_connections(self) -> List[str]:
    """Returns all known connections.

    Returns:
       Known connection paths.
    """
    known_connections = []

    for entry in self.persistent_configs["devices"].values():
      identifier = entry.get("console_port_name")
      if not identifier:
        # Use the IP address instead of the ADB port for ADB over IP
        identifier = typing.cast(str, entry["adb_serial"]).replace(":5555", "")
      known_connections.append(typing.cast(str, identifier))
    for entry in self.persistent_configs["other_devices"].values():
      identifier = entry.get("console_port_name")
      if not identifier:
        identifier = entry["hub_port_name"]
      known_connections.append(typing.cast(str, identifier))

    return known_connections

  def _detect_get_info(
      self, device_class: custom_types.Device, connection: str
  ) -> Tuple[str, custom_types.DeviceConfig, custom_types.DeviceConfig]:
    """Returns name, persistent and optional info from device communication.

    Note: Any errors raised will be caught in parent method.

    Args:
        device_class: device class with get_detection_info method.
        connection: path to communication

    Returns:
        (Name, dict of persistent props, dict of options props).
    """
    device_type = device_class.DEVICE_TYPE
    detect_file = self._get_detect_log_file(connection)
    device_config = {
        "persistent": {
            "console_port_name": connection,
            "device_type": device_type,
            "name": "\t{}_detect".format(device_type)
        },
        "options": {},
        "make_device_ready": "on"
    }
    logger.info("Getting info from communication port {} for {}".format(
        connection, device_type))

    if device_class.COMMUNICATION_TYPE == "PtyProcessComms":
      device_config["persistent"]["console_port_name"] = (
          pty_process_utils.get_launch_command(
              connection, **device_class.PTY_PROCESS_COMMAND_CONFIG))

    device = device_class(
        manager=self.manager_weakref(),
        device_config=device_config,
        log_directory=self.log_directory,
        log_file_name=detect_file)
    try:
      device.make_device_ready()
      persistent_props, options_props = device.get_detection_info()
    finally:
      device.close()

    name = self._generate_name(device_type, persistent_props["serial_number"],
                               device_class)
    persistent_props["name"] = name
    return name, persistent_props, options_props

  def _filter_out_known_connections(
      self, con_dict: Dict[str, List[str]], known_cons: List[str]
  ) -> Dict[str, List[str]]:
    """Filters out already known connections.

    Args:
        con_dict: dictionary of detected connections by
            connection class name as defined in communication_types.py
        known_cons: list of known connections paths.

    Returns:
        Filtered dictionary of unknown connections by connection class name.
    """
    new_con_dict = {}
    for key, con_list in con_dict.items():
      new_con_dict[key] = [
          con for con in con_list
          if con.replace(u":5555", u"") not in known_cons
      ]
      if new_con_dict[key]:
        logger.info("Found {} possible {} connections:".format(
            len(new_con_dict[key]),
            key.lower().replace("_", " ")))
        logger.info(u"\t" + u"\n\t".join(new_con_dict[key]))
    return new_con_dict

  def _generate_name(
      self,
      device_type: str,
      serial_number: str,
      device_class: custom_types.Device) -> str:
    """Generates name from last four digits of serial_number.

    Note: If there is a conflict with a different device it will generate a
        longer name.

    Args:
        device_type: device type.
        serial_number: device serial number.
        device_class: identifying class

    Returns:
        str: name of device

    Raises:
        RuntimeError: if serial number is too short.
    """
    if len(serial_number) < 4:
      raise RuntimeError("{} serial number {} is too short.".format(
          device_type, serial_number))
    name = "{}-{}".format(device_type, serial_number[-4:].lower())

    # Check if name already exists
    if device_class in self.auxiliary_classes:
      key = "other_devices"
    else:
      key = "devices"
    if name in self.persistent_configs[key]:
      # Check if its the same device or not
      if serial_number != self.persistent_configs[key][name]["serial_number"]:
        name = "{}-{}".format(device_type, serial_number[-8:].lower())
    return name

  def _get_detect_log_file(self, address: str) -> str:
    name = address.replace("/", "_")
    return "{}_detect.txt".format(name)

  def _identify_connection_device_class(
      self, connections_dict: Dict[str, List[str]]
  ) -> Tuple[
      List[Tuple[custom_types.Device, str]],
      List[str],
      List[str]]:
    """Detects all connections' appropriate device class.

    Args:
        connections_dict: dictionary of connections by communication type

    Returns:
        (List of (class, connection), errs, connections). not id'd
    """
    possible_device_tuples = []
    errs = []
    no_id_cons = []
    logger.info("\n##### Step 2/3 Identify Device Type of Connections. #####\n")

    for key in sorted(connections_dict.keys()):
      if not connections_dict[key]:  # no connections of that type
        continue
      key_name = key.lower().replace("_", " ")
      logger.info("Identifying {} devices..".format(key_name))
      for connection in connections_dict[key]:
        detect_log = os.path.join(self.log_directory,
                                  self._get_detect_log_file(connection))
        matching_classes = detect_criteria.determine_device_class(
            connection,
            key,
            detect_log,
            # pytype: disable=attribute-error
            self.manager_weakref().create_switchboard
            # pytype: enable=attribute-error
        )
        if len(matching_classes) > 1:
          device_types = [
              device_class.DEVICE_TYPE for device_class in matching_classes
          ]
          logger.warning(
              "Warning: Multiple device classes matched connection {}. "
              "This is a bug in the registered extension packages ({}). "
              "Returning {}.".format(device_types,
                                     extensions.get_registered_package_info(),
                                     device_types[0]))
        if matching_classes:
          logger.info("\t{} is a {}.".format(connection,
                                             matching_classes[0].DEVICE_TYPE))
          possible_device_tuples.append((matching_classes[0], connection))
        else:
          logger.info(
              "\t{} responses did not match a known device type.".format(
                  connection))
          no_id_cons.append(connection)
      logger.info("\t{} device_type detection complete.".format(key_name))
    return possible_device_tuples, errs, no_id_cons

  def _print_summary(
      self, names: List[str], errs: List[str], no_id_cons: List[str]) -> None:
    """Prints summary of detection events.

    Args:
        names: list of detected names.
        errs: list of errors.
        no_id_cons: list of undetected connections.
    """
    logger.info("\n##### Detection Summary #####\n")
    logger.info("\t{} new devices detected:{}", len(names),
                "\n\t\t" + "\n\t\t".join(names))
    if errs:
      logger.info("\n\t{} errors/warnings:{}", len(errs),
                  "\n\t\t" + "\n\t\t".join(errs))
    if no_id_cons:
      logger.info("\n\t{} connections found but not detected:{}",
                  len(no_id_cons), "\n\t\t" + "\n\t\t".join(no_id_cons))
    if errs or no_id_cons:
      logger.info("\nIf a connection failed detection, check {} for tips\n",
                  WIKI_URL)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Build Flasher capability interface."""
import abc
from typing import Any, Dict
from gazoo_device.capabilities.interfaces import capability_base
UNKNOWN = "Unknown"


class FlashBuildBase(capability_base.CapabilityBase):
  """Abstract base class for the build_flasher capability."""

  @abc.abstractmethod
  def download_build_file(self, remote_build_folder, local_folder) -> None:
    """Retrieves the build file(s) from the remote location and puts them in the local folder.

    Args:
        remote_build_folder (str): build folder in remote location.
        local_folder (str): local path to store build files.
    """

  @abc.abstractmethod
  def extract_build_info(self, build_args=None) -> Dict[str, str]:
    """Converts the provided build arguments into information about the build.

    Args:
        build_args (dict): dictionary of build arguments. If None, the
          default build arguments are used.

    Returns:
        dict: dictionary of build information including
        absolute_remote_path, build_type, and build_version.
    """

  @abc.abstractmethod
  def flash_device(self,
                   list_of_files,
                   expected_version=None,
                   expected_build_type=None,
                   verify_flash=True,
                   method=None):
    """Transfer a list of local files onto the device and flash the build on the device.

    Args:
        list_of_files (list): list of files on local computer.
        expected_version (str): firmware_version that is expected for the
          given list of files.
        expected_build_type (str): build_type that is expected for the given
          list of files.
        verify_flash (bool): True if flash_device should verify the
          firmware_version and build_type following the upgrade.
        method (str): upgrade method.

    Note: The list of files and expected version will be used to do an over
        the wire upgrade for the device.
    """

  @abc.abstractmethod
  def get_defaults(self) -> Dict[str, Any]:
    """Returns a dictionary of default build arguments.

    Returns:
        dict: the default build arguments for the device.
    """

  @abc.abstractmethod
  def get_firmware_type(self, build_args=None) -> str:
    """Returns the firmware type of the requested build.

    Args:
        build_args (dict): dictionary of build arguments. If None, the
          default build arguments are used.

    Returns:
        str: firmware type
    """

  @abc.abstractmethod
  def get_firmware_version(self, build_args=None) -> str:
    """Returns the firmware version based on the build arguments.

    Args:
        build_args (dict): dictionary of build arguments. If None, the
          default build arguments are used.

    Returns:
        str: firmware version
    """

  @abc.abstractmethod
  def get_remote_build_folder(self, build_args=None) -> str:
    """Uses the build arguments to determine the remote build folder.

    Args:
        build_args (dict): dictionary of build arguments. If None, the
          default build arguments are used.

    Returns:
        str: the remote build folder path.
    """

  @abc.abstractmethod
  def latest_verified_build_folder(self):
    """Returns the remote build folder path for the latest verified build.

    Returns:
        str: The latest verified remote build folder path.
    """

  @abc.abstractmethod
  def upgrade(self,
              build_number=None,
              build_url=None,
              build_file=None,
              forced_upgrade=False,
              latest_verified=False,
              **other_build_args):
    """Upgrade the device based on the provided build arguments.

    Args:
        build_number(int): build number. Defaults to last_believed good
          build.
        build_url(str): URL or gs:// path to the .zip upgrade file.
        build_file(str): local path to the file.
        forced_upgrade(bool): Upgrade the firmware without checks.
        latest_verified(bool): Upgrade to latest verified build.
        **other_build_args(dict): Other build arguments that are different
          based on the device type.
    """

  @abc.abstractmethod
  def upgrade_over_the_wire(self, **build_args):
    """Using the build arguments, download and flash the build on the device.

    Args:
        **build_args (dict): dictionary of build arguments.
    """



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Configuration for the interactive console (console.py)."""
import dataclasses
import re
from typing import Callable, Sequence, Tuple, Type, Union

from gazoo_device.switchboard import line_identifier
from gazoo_device.switchboard import log_process


@dataclasses.dataclass(frozen=True)
class _Window:
  """Configuration for a single console window."""
  name: str  # Window title.
  window_id: int  # Unique ID.


@dataclasses.dataclass(frozen=True)
class CommandResponseWindow(_Window):
  """Configuration for a command & response console window."""


@dataclasses.dataclass(frozen=True)
class LogWindow(_Window):
  """Configuration for a log console window."""


@dataclasses.dataclass(frozen=True)
class CommandInputWindow(_Window):
  """Configuration for a command input console window."""
  # Transport number to send commands to.
  transport: int
  # If True, handle Ctrl-C by reopening the transport instead of sending Ctrl-C
  # to the device as a workaround for ADB and SSH ignoring Ctrl-C keystrokes in
  # non-interactive mode (b/194440950).
  reopen_transport_on_ctrl_c: bool = False

WindowType = Union[CommandResponseWindow, LogWindow, CommandInputWindow]

_DEFAULT_LINE_TRANSFORMS = (  # (Before_regex, replacement string) tuples.
    ("GDM-M: Note", "GDM"),
    (r"GDM-\d: ", ""),
)
_LOG_LINE_PREFIX_LENGTH = (
    log_process.LOG_LINE_HEADER_LENGTH + log_process.HOST_TIMESTAMP_LENGTH)


@dataclasses.dataclass(frozen=True)
class ConsoleConfiguration:
  """Interactive console configuration."""
  # Console window arrangement. Windows are stacked vertically with the first
  # sequence element rendered at the top.
  windows: Sequence[WindowType]
  # Function which accepts a log line and returns the ID of the console window
  # to which this log line should be appended.
  line_to_window_id: Callable[[str], int]
  # Transformations to apply to each line. Transformations are applied in order.
  # Each transformation is defined by a regex pattern and a corresponding
  # replacement string.
  window_line_transforms: Sequence[Tuple[str, str]] = _DEFAULT_LINE_TRANSFORMS

  def __post_init__(self) -> None:
    """Validates the provided window arrangement.

    Raises:
      ValueError: If window IDs are not unique.
    """
    ids = [window.window_id for window in self.windows]
    if len(set(ids)) != len(ids):
      raise ValueError(f"Window IDs {ids} are not unique")


def calculate_window_id(
    transport_num: int, window_class: Type[WindowType]) -> int:
  """Returns a unique window ID.

  window_id numbering scheme does not depend on the number of windows present.
  This allows us to reuse the same line_to_window_id logic for window
  configurations with different numbers of windows.

  Args:
    transport_num: Switchboard transport number.
    window_class: Type of the window, which determines its offset.
  """
  max_windows_per_transport = 3
  if window_class is CommandInputWindow:
    offset = 0
  elif window_class is CommandResponseWindow:
    offset = 1
  else:  # LogWindow
    offset = 2
  return max_windows_per_transport * transport_num + offset


def make_line_to_window_id(
    windows: Sequence[WindowType],
    device_line_identifier: line_identifier.LineIdentifier
) -> Callable[[str], int]:
  """Creates a line_to_window_id callable for the given configuration."""
  first_non_input_window_id = next(
      window.window_id for window in windows
      if not isinstance(window, CommandInputWindow))

  def line_to_window_id(log_or_command_line: str) -> int:
    """Returns the window ID of the window which should contain the line."""
    port_match = re.search(r" GDM-(\d): ", log_or_command_line)
    if not port_match:  # It's an "GDM-M" line.
      send_port_match = re.search(
          r"wrote command.*to port (\d)", log_or_command_line)
      if send_port_match:
        port = int(send_port_match.group(1))
        window_id = calculate_window_id(port, CommandResponseWindow)
      else:
        window_id = first_non_input_window_id
    else:
      port = int(port_match.group(1))
      line = log_or_command_line[_LOG_LINE_PREFIX_LENGTH:]
      is_response = device_line_identifier.accept(
          port, line, line_identifier.LINE_TYPE_RESPONSE)
      is_log = device_line_identifier.accept(
          port, line, line_identifier.LINE_TYPE_LOG)
      if is_response:
        window_type = CommandResponseWindow
      elif is_log:
        window_type = LogWindow
      else:  # Unidentified lines are typically echoed inputs.
        window_type = CommandResponseWindow
      window_id = calculate_window_id(port, window_type)
    return window_id

  return line_to_window_id


# Helper functions for creation of common console configurations.
def get_log_response_separate_port_configuration(
    device_line_identifier: line_identifier.LineIdentifier,
    input_window_name: str = "Command input",
    response_window_name: str = "Commands & responses",
    log_window_name: str = "Logs",
    reopen_command_transport_on_ctrl_c: bool = True) -> ConsoleConfiguration:
  """Returns a configuration where port 0 is for commands and 1 is for logs."""
  windows = (
      CommandInputWindow(
          name=input_window_name,
          window_id=calculate_window_id(0, CommandInputWindow),
          transport=0,
          reopen_transport_on_ctrl_c=reopen_command_transport_on_ctrl_c),
      CommandResponseWindow(
          name=response_window_name,
          window_id=calculate_window_id(0, CommandResponseWindow)),
      LogWindow(
          name=log_window_name,
          window_id=calculate_window_id(1, LogWindow)),
  )
  return ConsoleConfiguration(
      windows=windows,
      line_to_window_id=make_line_to_window_id(windows, device_line_identifier),
  )


def get_log_response_same_port_configuration(
    device_line_identifier: line_identifier.LineIdentifier,
    input_window_name: str = "Command input (port 0)",
    response_window_name: str = "Commands & responses (port 0)",
    log_window_name: str = "Logs (port 0)",
    enable_secondary: bool = False,
    secondary_input_window_name: str = "Command input (port 1)",
    secondary_response_window_name: str = "Commands & responses (port 1)",
    secondary_log_window_name: str = "Logs (port 1)") -> ConsoleConfiguration:
  """Returns a configuration where each port is used for logs and responses."""
  windows = (
      CommandInputWindow(
          name=input_window_name,
          window_id=calculate_window_id(0, CommandInputWindow),
          transport=0),
      CommandResponseWindow(
          name=response_window_name,
          window_id=calculate_window_id(0, CommandResponseWindow)),
      LogWindow(
          name=log_window_name,
          window_id=calculate_window_id(0, LogWindow)),
  )
  if enable_secondary:
    windows += (
        CommandInputWindow(
            name=secondary_input_window_name,
            window_id=calculate_window_id(1, CommandInputWindow),
            transport=1),
        CommandResponseWindow(
            name=secondary_response_window_name,
            window_id=calculate_window_id(1, CommandResponseWindow)),
        LogWindow(
            name=secondary_log_window_name,
            window_id=calculate_window_id(1, LogWindow)),
    )

  return ConsoleConfiguration(
      windows=windows,
      line_to_window_id=make_line_to_window_id(windows, device_line_identifier),
  )


def get_log_only_configuration(
    log_window_name: str = "Logs") -> ConsoleConfiguration:
  """Returns a configuration where only logs from port 0 are shown."""
  log_window_id = calculate_window_id(0, LogWindow)
  return ConsoleConfiguration(
      windows=(
          LogWindow(name=log_window_name, window_id=log_window_id),
      ),
      line_to_window_id=lambda line: log_window_id,  # There's only 1 window.
  )




# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Interface for an PwRPC (Pigweed RPC) echo capability."""
import abc
from gazoo_device.capabilities.interfaces import capability_base


class PwRPCEchoBase(capability_base.CapabilityBase):
  """Pigweed RPC echo capability."""

  @abc.abstractmethod
  def echo(self, msg: str) -> str:
    """Sends a message to the echo endpoint and returns the response back."""



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Interface for an PwRPC (Pigweed RPC) lighting capability."""
import abc
from gazoo_device.capabilities.interfaces import capability_base
from gazoo_device.protos import lighting_service_pb2


class PwRPCLightBase(capability_base.CapabilityBase):
  """Pigweed RPC lighting capability for devices communicating over PwRPC."""

  @abc.abstractmethod
  def on(self, no_wait: bool = False) -> None:
    """Turns on the light state of the device.

    Args:
      no_wait: If True, returns before verifying the light state.

    Raises:
      DeviceError: When the device does not transition to the appropriate
      state or if it remains off.
    """

  @abc.abstractmethod
  def off(self, no_wait: bool = False) -> None:
    """Turns off the light state of the device.

    Args:
      no_wait: If True, returns before verifying the light state.

    Raises:
      DeviceError: When the device does not transition to the appropriate
      state or if it remains on.
    """

  @property
  @abc.abstractmethod
  def state(self) -> bool:
    """The light state of the device.

    Returns:
      True if the device is in on state, false if it's in off state.
    """

  @property
  @abc.abstractmethod
  def brightness(self) -> int:
    """The brightness level of the device: between 0 and 255 inclusive.

    Returns:
      The current brightness level.
    """

  @property
  @abc.abstractmethod
  def color(self) -> lighting_service_pb2.LightingColor:
    """The lighting color of the device.

    Color consists of hue and saturation, which are between 0x00 and 0xFE
    inclusive.

    Returns:
      The current lighting color.
    """



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Common shell() capability for devices communicating over SSH."""
import time
from gazoo_device import config
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.capabilities.interfaces import shell_base

_SSH_CONNECTION_FAILURE_MARKERS = ["Connection to", "Connection reset"]

logger = gdm_logger.get_logger()


class ShellSSH(shell_base.ShellBase):
  """Common shell() method implementation for devices communicating over SSH."""

  def __init__(self,
               send_and_expect,
               device_name,
               shell_cmd=shell_base._SHELL_CMD,
               shell_regex=shell_base._SHELL_REGEX,
               tries=shell_base._TRIES,
               timeout=shell_base._TIMEOUT,
               failure_markers=None):
    """Initalize the SSH shell capability.

    Args:
        send_and_expect (method): bound send_and_expect method of the device
          class instance.
        device_name (str): name of the device using this capability.
        shell_cmd (str): return code wrapper around the shell command to
          execute.
        shell_regex (str): shell regex to use. Must contain two capturing
          groups: one for the output and one for the return code.
        tries (int): how many times to try sending the shell command.
        timeout (float): shell timeout in seconds.
        failure_markers (list): list of markers (strings) indicating SSH
          connection failure.
    """
    super(ShellSSH, self).__init__(
        send_and_expect=send_and_expect,
        device_name=device_name,
        shell_cmd=shell_cmd,
        shell_regex=shell_regex,
        tries=tries,
        timeout=timeout)

    if failure_markers is None:
      failure_markers = _SSH_CONNECTION_FAILURE_MARKERS

    self._failure_markers = failure_markers

  def shell(self,
            command,
            command_name="shell",
            timeout=None,
            port=0,
            include_return_code=False,
            searchwindowsize=config.SEARCHWINDOWSIZE):
    """Sends command and returns response and optionally return code.

    Args:
        command (str): Command to send to the device.
        command_name (str): Identifier for command.
        timeout (float): Time in seconds to wait for device to respond.
        port (int): Which port to send on, 0 or 1.
        include_return_code (bool): flag indicating return code should be
          returned.
        searchwindowsize (int): Number of the last bytes to look at.

    Raises:
        DeviceError: if communication fails.

    Note:
        Can try multiple times as connection can sometimes fail.
        See the init args for setting the number of retry attempts.

    Returns:
        str: If include_return_code is False return the device response to
        the command.
        tuple: If include_return_code is True return the device response and
        return code.
    """
    if timeout is None:
      timeout = self._timeout

    command = self._shell_cmd.format(cmd=command.rstrip())

    logger.debug("{} sending {!r} to generate {} in {}s on port {}",
                 self._device_name, command, command_name, timeout, port)

    for attempt in range(self._tries):
      response = self._send_and_expect(
          command, [self._shell_regex],
          timeout=timeout,
          port=port,
          searchwindowsize=searchwindowsize,
          expect_type="response")
      if not response.timedout:
        break

      if (any(failure_marker in response.before
              for failure_marker in self._failure_markers) and
          attempt < self._tries - 1):
        logger.warning(
            "{}: SSH connection died with output {}. Trying again.".format(
                self._device_name, response.before))
        # SSH connection died. Retry.
        time.sleep(.1)
      else:
        raise errors.DeviceError("Device {} shell failed for command {!r}. "
                                 "Timed out waiting {}s for response. "
                                 "Shell output: {!r}.".format(
                                     self._device_name, command, timeout,
                                     response.before))

    result = response.match.group(1).strip()
    return_code = int(response.match.group(2))

    if include_return_code:
      return result, return_code
    else:
      return result

  def has_command(self, binary_name):
    """Returns if binary_name is installed on the device.

    Args:
        binary_name (str): name of the executable.

    Returns:
        bool: True if the executable is found on the device, False
        otherwise.
    """
    _, result_code = self.shell(
        f'which {binary_name}\n', include_return_code=True)
    return result_code == 0



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Log event parser capability interface.

Parser provides general purpose regular expression matching on a line by line
basis with the ability to extract both the last and history of matching events.

When the Parser class is created it loads filter files or filter file
directories specified in the `filters` argument.

JSON filter file format (version 1.0):
* A single JSON object with the following keys:

    * "description" whose value is a string (optional)
    * "version" whose value is a JSON object with the following keys:

        * "major" whose value is an integer that specifies the major version of
        the file
        * "minor" whose value is an integer that specifies the major version of
        the file

    * "filters" whose value is an array of Filter JSON objects with the
    following keys:

        * "description" whose value is a string (optional)
        * "regex_match" whose value is a regular expression string
        * "name" whose value is a string

::

    {
        "description": "helpful optional description for all filters contained",
        "version": {"major": 1, "minor": 0},
        "filters": [
            {
                "description": "optional description for this specific filter",
                "regex_match": "required regular expression string",
                "name": "required_name"
            }
        ]
    }

As each JSON filter file is loaded and parsed each regular expression is
assigned a name consisting of the JSON filter file name (without the extension)
plus the name assigned to each Filter object (the "name" key) separated with a
'.' (e.g. errors.assert). This is known in the Parser code base as the
"event_label" and is used as a lookup key to retrieve Parser filter events.

The Parser class expects to store all matching events in an event file. To
reduce the memory footprint of the Parser class all matching events are
immediately written to the event file by the process_line method. The caller
must provide the shared dynamic event dictionary, an event file object, and the
raw log line to this method. Keeping the shared dynamic event dictionary and the
opening and closing of the event file outside the Parser class helps to simplify
the Parser class logic and ensures that the file is closed by the caller as
needed (e.g. exception handling).

For each regular expression that matches the raw log line provided to the
process_line method the shared dynamic event dictionary is updated. When all
regular expressions have been processed an event JSON object containing the
results of all matches by "event_lable" is written (and flushed) to the event
file. The event JSON object format is shown as follows:

JSON event file format (version 1.0):
* A single JSON object with the following keys:

    * "raw_log_line" whose value is the raw log line string parsed
    * "system_timestamp" whose value is a string containing the timestamp
    portion of the
      raw log line
    * "event_label" for each regular expression that matched the log line and
    whose value is an
      array of regular expression group data extracted

::

    {
        "system_timestamp": "2018-01-30 12:05:59.118746",
        "state.shutdown_reboot": ["The system is shutting down"],
        "raw_log_line": "<2018-01-30 12:05:59.118746> announce: The system is
        shutting down."
    }

Since event history can only be obtained from the event file, all event history
commands require the path to the event file to retrieve this information. Event
history is obtained by using the "tac" and "grep" unix tools to filter matching
events in the event file and from Python extract and decode each event JSON
object as described above.
"""
import abc
from gazoo_device.capabilities.interfaces import capability_base

LABEL_REBOOT_TRIGGER = "basic.reboot_trigger",
LABEL_BOOTUP = "basic.bootup"


class ParserResult(object):
  """Store the results of a Parser API call.

    Attributes:
        timedout (bool): True if the Parser request timed out; False otherwise
        results_list (list): A list of Event Results
        count (int): The number of Event Results returned

    Notes:
        An Event Result is a dict with the following format:
            {"system_timestamp": datetime,
             "raw_log_line": str,
             event_label: list, ...}
        The value of the "system_timestamp" key is
        a datetime object with the timestamp when the log line was
        captured by GDM. The value for the "raw_log_line" key is a string
        containing the entire log line captured by GDM, not including the
        timestamp added by GDM. The Event Result also contains one or
        more key - value pairs representing event matches for that log
        line, i.e., there will always be at least one event match, but
        there may be more than one if multiple event regexes match the
        same line.  For each match, the key is the event_label, and the
        value is a list of group matches for the corresponding regex. If
        no groups were specified in the event regex, the group matches
        list will be empty.
  """

  def __init__(self, timedout, results_list, count):
    self.timedout = timedout
    self.results_list = results_list
    self.count = count


class EventParserBase(capability_base.CapabilityBase):
  """Log event parser capability interface."""

  @abc.abstractmethod
  def get_event_history(self, event_labels=None, count=None, timeout=10.0):
    r"""Returns up to count elements of event data matching given list of event labels.

    Args:
        event_labels (list): list of event labels to lookup(example:
          ["power.state", "power.lost"]). If None(default), return all
          events.
        count (int): of event data elements to return (Default: None for all
          events).
        timeout (float): Timeout value in seconds. Example: 10.0.

    Raises:
        ParserError: on error parsing event file.

    Returns:
        ParserResult: An object containing events matching any of the event
        labels provided in event_labels, or all available events if event_labels
        in None. The event data is stored in the .results_list attribute and the
        count of results in the .count attribute. If no matching events are
        found, the .results_list attribute will be set to the empty list.

        If the .timedout attribute is True, the operation timed out before all
        events were searched, or the event file was not found before the timeout
        expired.

    Notes:
        Event info is listed in most recent to least recent order.

        The following examples use a hypothetical power.json filter file
        with the following filters:

        .. code-block:: none

            {
                "name": "state",
                "regex_match": "power:(\\w+)"
            },
            {
                "name": "lost",
                "regex_match": "power lost"
            },
            {
                "name": "restored",
                "regex_match": "power restored"
            }

        The original log file being read looks like this:

        .. code-block:: none

            <2018-02-02 10:32:07.167234> [APPL] Spoke: power restored,
            powering on
            <2018-02-02 12:00:57.154328> [APPL] Spoke: power lost, powering
            down
            <2018-02-02 12:00:57.154328> power:OFF
            <2018-02-02 12:01:22.122824> [APPL] Spoke: power restored,
            powering on
            <2018-02-02 12:01:22.122824> power:ON
            <2018-02-02 12:02:18.121948> [APPL] Spoke: power lost, powering
            down
            <2018-02-02 12:02:18.121948> power:OF

        Example output with event_labels ["power.state"], showing group
        matches:

        .. code-block:: none

            result.timedout = False
            result.results_list =
                [{'system_timestamp': datetime.datetime(2018, 2, 2, 12, 2,
                18, 121948),
                  'raw_log_line': 'power:OFF',
                  'power.state': ['OFF']},
                 {'system_timestamp': datetime.datetime(2018, 2, 2, 12, 1,
                 22, 122824),
                  'raw_log_line': 'power:ON',
                  'power.state': ['ON']},
                 {'system_timestamp': datetime.datetime(2018, 2, 2, 12, 0,
                 57, 154328),
                  'raw_log_line': 'power:OFF',
                  'power.state': ['OFF']}]
            result.count = 3

        Example output with event_labels ["power.lost"], showing matches
        with no groups:

        .. code-block:: none

            result.timedout = False
            result.results_list =
                [{'system_timestamp': datetime.datetime(2018, 2, 2, 12, 2,
                18, 121948),
                  'raw_log_line: '[APPL] Spoke: power lost, powering down',
                  'power.lost': []},
                 {'system_timestamp': datetime.datetime(2018, 2, 2, 12, 0,
                 57, 154328),
                  'raw_log_line': '[APPL] Spoke: power lost, powering down',
                  'power.lost': []}]
            result.count = 2

        Example output with event_labels["power.restored"], showing matches
        with no groups and count=1:

        .. code-block:: none

            result.timedout = False
            result.results_list =
                [{'system_timestamp': datetime.datetime(2018, 2, 2, 12, 1,
                22, 122824),
                  'raw_log_line': '[APPL] Spoke: power restored, powering
                  on',
                  'power.restored': []}
            result.count = 1

        Example output with event_labels ["power.lost", "power.restored"]:

        .. code-block:: none

            result.timedout = False
            result.results_list =
                [{'system_timestamp': datetime.datetime(2018, 2, 2, 12, 2,
                18, 121948),
                  'raw_log_line: '[APPL] Spoke: power lost, powering down',
                  'power.lost': []},
                 {'system_timestamp': datetime.datetime(2018, 2, 2, 12, 1,
                 22, 122824),
                  'raw_log_line': '[APPL] Spoke: power restored, powering
                  on',
                  'power.restored': []},
                 {'system_timestamp': datetime.datetime(2018, 2, 2, 12, 0,
                 57, 154328),
                  'raw_log_line': '[APPL] Spoke: power lost, powering down',
                  'power.lost': []}
                 {'system_timestamp': datetime.datetime(2018, 2, 2, 10, 32,
                 7, 167234),
                  'raw_log_line': '[APPL] Spoke: power restored, powering
                  on',
                  'power.restored': []}]
            result.count = 4
    """

  @abc.abstractmethod
  def get_event_history_count(self, event_label, timeout=10.0):
    """Returns count of event history matching event label specified.

    Args:
        event_label(str): event label to lookup(e.g. power.state)
        timeout(float): Timeout value in seconds. Example: 10.0.

    Raises:
        ParserError: on error parsing event file

    Returns:
        ParserResult: An object containing the count of results in the
        .count attribute. The .results_list field is set to the empty list.

        If the .timedout attribute is True, the operation timed out before
        all events were searched, or the event file was not found before the
        timeout expired.
    """

  @abc.abstractmethod
  def get_event_label_dict(self, pattern=None):
    """Takes an event label pattern, and returns a dictionary of matching event labels.

    Args:
        pattern(str): Regex for event labels. Default: None

    Returns:
        dict: containing keys(i.e. event labels), and values(i.e regular
        expressions)

    Raises:
        ParserError: if pattern provided is an invalid regular expression

    Note:
        The pattern regular expression provided will be matched strictly
        against the beginning of each event_label.
        If no pattern is provided, the _filters_dict is returned.
    """

  @abc.abstractmethod
  def get_event_labels(self, pattern=None):
    """Takes an event label pattern, and returns a list of matching event labels.

    Args:
        pattern(str):  Regex for event labels. Default: None

    Returns:
        list: of matching event label strings

    Raises:
        ParserError: if pattern provided is an invalid regular expression

    Note:
        The pattern regular expression provided will be matched strictly
        against the beginning of each event_label.
        If no pattern is provided, all filter event labels are returned.
    """

  @abc.abstractmethod
  def get_last_event(self, event_labels=None, timeout=1.0):
    r"""Returns the most recent matching event for each item in the list of event labels.

    Args:
        event_labels (list): list of event labels to lookup (example:
          ["power.state", "power.lost"]). None (default) returns the most
          recent event regardless of label.
        timeout (float): Timeout value in seconds. Example: 1.5.

    Raises:
        ParserError: on error parsing event file.
        DeviceError: Label doesn't exist or
                     event_labels is not type list

    Returns:
        ParserResult: An object containing the most recent event match for
        each event label in the event_labels list, or the most recent event if
        event_labels is None. The event data is stored in the .results_list
        attribute, and the count of results is in the .count attribute. If
        no matching events are found, the .results_list attribute will be set to
        the empty list.

        If the .timedout attribute is True, the operation timed out before
        all events were searched, or the event file was not found before the
        timeout expired.

    Notes:
        The event_label is the desired filter filename without extension a .
        and the filter name (e.g. power.state).

        The following examples use a hypothetical power.json filter file
        with the following filters:

        .. code-block:: none

            {
                "name": "state",
                "regex_match": "power:(\\w+)"
            },
            {
                "name": "lost",
                "regex_match": "power lost"
            },
            {
                "name": "restored",
                "regex_match": "power restored"
            }

        The original log file being read looks like this:

        .. code-block:: none

            <2018-02-02 10:32:07.167234> [APPL] Spoke: power restored,
            powering on
            <2018-02-02 12:00:57.154328> [APPL] Spoke: power lost, powering
            down
            <2018-02-02 12:00:57.154328> power:OFF
            <2018-02-02 12:01:22.122824> [APPL] Spoke: power restored,
            powering on
            <2018-02-02 12:01:22.122824> power:ON
            <2018-02-02 12:02:18.121948> [APPL] Spoke: power lost, powering
            down
            <2018-02-02 12:02:18.121948> power:OFF

        Example output with event_labels=["power.state"], showing group
        matches:

        .. code-block:: none

            result.timedout = False
            result.results_list =
                [{'system_timestamp': datetime.datetime(2018, 2, 2, 12, 2,
                18, 121948),
                  'raw_log_line': 'power:OFF',
                  'power.state': ['OFF']}]
            result.count = 1

        Example output with event_labels=["power.lost"], showing matches
        with no groups:

        .. code-block:: none

            result.timedout = False
            result.results_list =
                [{'system_timestamp': datetime.datetime(2018, 2, 2, 12, 2,
                18, 121948),
                  'raw_log_line': '[APPL] Spoke: power lost',
                  'power.lost': []}]
            result.count = 1

        Example output with event_labels=["power.state", "power.lost",
        "power.restored"]:

        .. code-block:: none

            result.timedout = False
            result.results_list =
                [{'system_timestamp': datetime(2018, 2, 2, 12, 2, 18,
                121948),
                  'raw_log_line': 'power:OFF',
                  'power.state': ['OFF']},
                 {'system_timestamp': datetime(2018, 2, 2, 12, 2, 18,
                 121948),
                  'raw_log_line': '[APPL] Spoke: power lost',
                  'power.lost': []}
                 {'system_timestamp': datetime(2018, 2, 2, 12, 1, 22,
                 122824),
                  'raw_log_line': '[APPL] Spoke: power restored, powering
                  on',
                  'power.restored': []}]
            results.count = 3
    """

  @abc.abstractmethod
  def get_last_event_state(self,
                           event_label,
                           event_name=None,
                           raise_error=False,
                           timeout=1.0,
                           group_index=0):
    r"""Return the matching group of the event label regex if one is found.

    Args:
        event_label(str): The event label based on the filter name and
          filter file name.
        event_name (str): name of event. Defaults to event_label.
        raise_error(bool): Raise an exception if the event has not been seen
          or otherwise return an empty string.
        timeout (float): time to parse events.
        group_index (int): index of the match group to return (default 0 for
          first).

    Returns:
        str: The result from the call to get_last_event and getting the
        first group.

    Raises:
        ParserError: if there is an error getting the last event.

        .. code-block:: none
            {
                "name": "state",
                "regex_match": "power:(\\w+)"
            }

        The original log file being read looks like this:

        .. code-block:: none

            <2018-02-02 10:32:07.167234> [APPL] Spoke: power restored,
            powering on
            <2018-02-02 12:00:57.154328> [APPL] Spoke: power lost, powering
            down
            <2018-02-02 12:00:57.154328> power:OFF
            <2018-02-02 12:01:22.122824> [APPL] Spoke: power restored,
            powering on
            <2018-02-02 12:01:22.122824> power:ON
            <2018-02-02 12:02:18.121948> [APPL] Spoke: power lost, powering
            down
            <2018-02-02 12:02:18.121948> power:OFF

        With event_labels=["power.state"], it would return 'OFF'
    """

  @abc.abstractmethod
  def get_unexpected_reboots(self):
    """Device's event file is used to identify unmatched reboot events.

    Returns:
        list: unexpected reboot event

    Note:
        For example in the following case there are two reboot triggers but
        only one bootup event::

            {"basic.reboot_trigger": [],
             "system_timestamp": datetime.datetime.strptime("2018-06-18
             15:13:48.011747",
                                                            "%Y-%m-%d
                                                            %H:%M:%S.%f"),
             "raw_log_line": "Note: GDM triggered reboot"},
            {"basic.reboot_trigger": [],
             "system_timestamp": datetime.datetime.strptime("2018-06-18
             15:13:52.825231",
                                                            "%Y-%m-%d
                                                            %H:%M:%S.%f"),
             "raw_log_line": "Note: GDM triggered reboot"}
            {"basic.bootup": [],
             "system_timestamp": datetime.datetime.strptime("2018-06-18
             15:13:58.540417",
                                                            "%Y-%m-%d
                                                            %H:%M:%S.%f"),
             "raw_log_line": "[0:00:00.035] [APPL] Welcome to Lock App"}

        This method would return::

            [{"basic.reboot_trigger": [],
             "system_timestamp": datetime.datetime.strptime("2018-06-18
             15:13:48.011747",
                                                            "%Y-%m-%d
                                                            %H:%M:%S.%f"),
             "raw_log_line": "Note: GDM triggered reboot"}]
    """

  @abc.abstractmethod
  def get_unmatched_events(self,
                           event_cause_label=LABEL_REBOOT_TRIGGER,
                           event_effect_label=LABEL_BOOTUP):
    """Device's event file is used to identify unmatched events.

    Args:
        event_cause_label (str): Name of event label causing event
        event_effect_label (str): Name of event label signifying effect event

    Returns:
        tuple:  list of unmatched cause event dict entries
                list of unmatched effect event dict entries

    Note:
        For example in the following case there are two reboot triggers but
        only one
        bootup event::

            {"basic.reboot_trigger": [],
             "system_timestamp": datetime.datetime.strptime("2018-06-18
             15:13:48.011747",
                                                            "%Y-%m-%d
                                                            %H:%M:%S.%f"),
             "raw_log_line": "Note: GDM triggered reboot"},
            {"basic.reboot_trigger": [],
             "system_timestamp": datetime.datetime.strptime("2018-06-18
             15:13:52.825231",
                                                            "%Y-%m-%d
                                                            %H:%M:%S.%f"),
             "raw_log_line": "Note: GDM triggered reboot"}
            {"basic.bootup": [],
             "system_timestamp": datetime.datetime.strptime("2018-06-18
             15:13:58.540417",
                                                            "%Y-%m-%d
                                                            %H:%M:%S.%f"),
             "raw_log_line": "[0:00:00.035] [APPL] Welcome to Lock App"}

        This method would return the following tuple::

            ([{"basic.reboot_trigger": [],
             "system_timestamp": datetime.datetime.strptime("2018-06-18
             15:13:48.011747",
                                                            "%Y-%m-%d
                                                            %H:%M:%S.%f"),
             "raw_log_line": "Note: GDM triggered reboot"}],
            [])
    """

  @abc.abstractmethod
  def load_filter_file(self, filter_path):
    """Load JSON filter file specified.

    Args:
        filter_path (str): path to JSON filter file

    Raises:
        ParserError: when loading JSON filter file fails or on unexpected
        JSON key values

    Note:
        Only files ending in .json are loaded, others are skipped

        Example:
            {
                "description": "Assertion event filter",
                "version": {"major": 1, "minor": 0},
                "filters": [
                    {
                        "description": "match and extract assert message"
                        "regex_match": "ASSERT: (.+)",
                        "name": "state"
                    }
                ]
            }
    """

  @abc.abstractmethod
  def process_line(self,
                   event_file,
                   raw_log_line,
                   header_length=29,
                   log_filename=""):
    """Parser line against filters and write JSON result to event_file provided.

    Args:
        event_file (file): to write matching filter results to
        raw_log_line (str): to match filters against
        header_length (int): added by GDM to strip off from raw_log_line
        log_filename (str): name of log file raw_log_line came from

    Note: The expected format for the raw_log_line is one that contains the
      system timestamp between characters 1 and 27 as shown in the following
      example:
        <2018-02-01 14:39:00.492376> ...
        A line containing a event_entry should be written as follows:
            * Key: log_filename, Value: filename where raw_log_line was found
            * Key: matched_timestamp, Value: a host timestamp when match was
              found.
            * Key: raw_log_line, Value: the line provided without any trailing
              whitespace
            * Key: system_timestamp, Value: the timestamp portion of the raw log
              line
            * Key: <event_label>, Value: regular expression group tuple as
              an array which might be empty  In addition to writing results
              to the event file the <event_label> is used as a key to update
              the shared dynamic events dictionary with the same information
              to allow for easier access to the data for the last matching
              <event_label> in the main process
    """

  @abc.abstractmethod
  def verify_event_labels(self, event_labels, error_message=""):
    """Verifies event_labels in correct format and exist.

    Args:
        event_labels (list): list of event_labels.
        error_message (str): error_message to use as header.

    Raises:
        ParserError: if bad format or certain labels don't exist.
    """

  @abc.abstractmethod
  def wait_for_event_labels(self,
                            event_labels,
                            raise_error=False,
                            timeout=20.0,
                            start_datetime=None):
    """Waits up to timeout seconds for event labels to appear in device logs.

    Args:
        event_labels (list): labels(strings) to wait for. Will wait for at
          least one of each.
        raise_error (bool): raise an error if the labels are not found in time.
        timeout (float): seconds to wait for labels to appear in logs.
        start_datetime (datetime): events before this time will be ignored.

    Raises:
        ParserError: if input format is bad
        ParserError: raise_error is True and not all labels were found.

    Returns:
        bool: whether all event labels were found.
    """

  @abc.abstractmethod
  def load_filters(self, filters):
    """Load JSON filter files or directories specified.

    Args:
        filters(list): of JSON filter files or paths to filter files

    Raises:
        ParserError: if files or paths do not exist

    Note:
        Only files ending in .json are loaded, others are skipped(see
        load_filter_file method for an example)
    """



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""EFR32 dev board device class."""
from gazoo_device import detect_criteria
from gazoo_device.base_classes import silabs_efr32_device
from gazoo_device.utility import pwrpc_utils


class EFR32(silabs_efr32_device.SilabsEFR32Device):
  """Base Class for EFR32 Devices.

  EFR32MG which includes Zigbee and Thread software stacks to serve as a
  CHIP-compliant device, see more details in
  https://www.silabs.com/documents/public/user-guides/ug342-brd4170a-user-guide.pdf
  """
  DETECT_MATCH_CRITERIA = {
      detect_criteria.PigweedQuery.product_name: "j-link",
      detect_criteria.PigweedQuery.manufacturer_name: r"silicon(_| )labs",
      detect_criteria.PigweedQuery.app_type:
          pwrpc_utils.PigweedAppType.NON_PIGWEED.value,
  }
  DEVICE_TYPE = "efr32"
  _OWNER_EMAIL = "gdm-authors@google.com"




# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implementation of the switch_power_ethernet."""
from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.capabilities.interfaces import switch_power_base
from gazoo_device.utility import common_utils

logger = gdm_logger.get_logger()

ON = "on"
OFF = "off"

TIMEOUTS = {"STATE_CHANGE": 5}


class SwitchPowerEthernet(switch_power_base.SwitchPowerBase):
  """Switch power flavor for ethernet switches."""

  def __init__(self, device_name, get_port_status_func, turn_on_port_func,
               turn_off_port_func):
    """Create an instance of the switch power with ethernet capability.

    Args:
        device_name (str): name, usually the switch ip address.
        get_port_status_func (func): function to get port status.
        turn_on_port_func (func): function to turn on port.
        turn_off_port_func (func): function to turn off port.
    """
    super().__init__(device_name=device_name)
    self._get_port_status_func = get_port_status_func
    self._turn_on_port_func = turn_on_port_func
    self._turn_off_port_func = turn_off_port_func

  @decorators.PersistentProperty
  def supported_modes(self):
    """Get the power modes supported by the switch."""
    return [OFF, ON]

  def get_mode(self, port):
    """Gets the mode for the specified port.

    Args:
      port (int): Use this port to get the mode.

    Returns:
        str: port mode settings 'off', 'on'

    Raises:
      DeviceError: invalid port.
    """
    port_status = self._get_port_status_func(port)
    if port_status.lower() in ["on", "1", "enable"]:
      return ON
    return OFF

  @decorators.CapabilityLogDecorator(logger, decorators.DEBUG)
  def power_off(self, port):
    """This command powers off the port specified.

    Args:
        port (int): Identifies which hub port to power off.
    """
    self.set_mode(OFF, port)

  @decorators.CapabilityLogDecorator(logger, decorators.DEBUG)
  def power_on(self, port):
    """This command powers on the port specified.

    Args:
        port (int): Identifies which hub port to power on.
    """
    self.set_mode(ON, port)

  @decorators.CapabilityLogDecorator(logger, decorators.DEBUG)
  def set_mode(self, mode, port):
    """Sets the given port to the mode specified.

    Args:
      mode (str): mode to set. Options: 'off', 'on'
      port (int): The port to set.

    Raises:
      DeviceError: invalid port, or mode.
    """
    self._validate_mode(mode)
    logger.info("{} setting power mode to {} for port {}".format(
        self._device_name, mode, port))

    if mode == ON:
      self._turn_on_port_func(port)
    else:
      self._turn_off_port_func(port)
    try:
      common_utils.retry(
          func=self._verify_mode_change,
          func_args=[port, mode],
          timeout=TIMEOUTS["STATE_CHANGE"],
          is_successful=common_utils.is_true,
          interval=1,
          reraise=False)
    except errors.CommunicationTimeoutError:
      raise errors.DeviceError(
          "{} failed to verify that ethernet connection is {}".format(
              self._device_name, mode))

  def _validate_mode(self, mode):
    if mode not in list(self.supported_modes):
      raise errors.DeviceError(
          "Mode {} not found in supported modes {!r}".format(
              mode, self.supported_modes))

  def _verify_mode_change(self, port, expected_mode):
    return self.get_mode(port) == expected_mode



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Interactive console to replace direct device I/O such as ADB, SSH, Kermit."""
import asyncio
import collections
import re
from typing import Optional, Sequence, Tuple

from gazoo_device import console_config
from gazoo_device import gdm_logger
from gazoo_device.capabilities.interfaces import switchboard_base
from prompt_toolkit import application
from prompt_toolkit import buffer
from prompt_toolkit import document
from prompt_toolkit import enums
from prompt_toolkit import filters
from prompt_toolkit import input as prompt_toolkit_input
from prompt_toolkit import key_binding
from prompt_toolkit import output
from prompt_toolkit import styles
from prompt_toolkit import widgets
from prompt_toolkit.application import current as application_current
from prompt_toolkit.eventloop import inputhook
from prompt_toolkit.key_binding.bindings import focus
from prompt_toolkit.layout import containers
from prompt_toolkit.layout import layout

logger = gdm_logger.get_logger()

_HELP_TEXT = """Application-level controls:
  Tab: move focus to the next window.
  Shift-Tab: move focus to the previous window.
  Mouse selection: select text to copy.
  Cmd-C/Win-C: copy selected text to the system clipboard.

Menu navigation controls:
  Use arrow keys to move between menu items.
  Use Enter to select the current menu item.

To close the console:
  Select the "Exit" menu button.

Command input window keys:
  Enter: send the contents of the input window to the device.
  Ctrl-C: clear the input window and interrupt the currently running command.
  Up/down arrow: select a command from history.
  Cmd-V/Win-V: paste system clipboard contents.
  Other keys (Left/Right Arrow, Page Up/Down, Backspace, Delete, characters):
    same as a typical text editor.

Command response and log window keys:
  Cursor controls:
    Arrow keys: move the cursor.
    vi-style key bindings can also be used to move the cursor. For example:
      gg: go to the beginning of text.
      G: go to the end of text.
    TODO(b/194440944): add an option for Emacs key bindings.

  Search controls:
    /: open forward search prompt.
    ?: open backward search prompt.
    Enter: perform the search.
    Esc: close the search prompt.
    n: go to the next match.
    N: go to the previous match.

Press Enter to close this window."""
_HELP_TEXT_WIDTH = 80
_MAX_TEXT_AREA_LINES = 100000


def _do_exit() -> None:
  application_current.get_app().exit()


class _AppendableTextArea(widgets.TextArea):
  """TextArea which allows appending text and automatically scrolls down."""

  def append_text(self, text_to_append: str) -> None:
    """Appends text_to_append to the text area."""
    new_text = self.text + text_to_append
    if not self.document.on_last_line:
      # Do not scroll down to latest log line if user is searching through logs.
      new_cursor_position = self.document.cursor_position
    else:
      new_cursor_position = len(new_text)
    self.document = document.Document(new_text, new_cursor_position)
    self.truncate_text()

  def truncate_text(self) -> None:
    """Drops the least recent lines if text size exceeds the maximum limit."""
    if self.document.line_count > _MAX_TEXT_AREA_LINES:
      num_lines_to_drop = self.document.line_count - _MAX_TEXT_AREA_LINES
      new_text = "\n".join(self.document.lines[num_lines_to_drop:])
      cursor_offset = (
          sum(len(line) for line in self.document.lines[:num_lines_to_drop])
          + num_lines_to_drop)  # To account for newline characters.
      if self.document.cursor_position >= cursor_offset:
        new_cursor_position = self.document.cursor_position - cursor_offset
      else:
        # If dropping lines at the current cursor position, imitate cursor
        # motion from pressing "down" several times.
        num_down_presses = num_lines_to_drop - self.document.cursor_position_row
        new_cursor_position = (
            self.document.cursor_position
            + self.document.get_cursor_down_position(num_down_presses)
            - cursor_offset)
      self.document = document.Document(new_text, new_cursor_position)


class _HelpWindowVisibility:
  """Controls for help window visibility."""

  def __init__(
      self,
      focus_element_on_show: Optional[containers.AnyContainer] = None,
      focus_element_on_hide: Optional[containers.AnyContainer] = None) -> None:
    self._visible = False
    self.focus_element_on_show = focus_element_on_show
    self.focus_element_on_hide = focus_element_on_hide

  def show(self) -> None:
    self._visible = True
    if self.focus_element_on_show:
      application_current.get_app().layout.focus(self.focus_element_on_show)

  def hide(self) -> None:
    self._visible = False
    if self.focus_element_on_hide:
      application_current.get_app().layout.focus(self.focus_element_on_hide)

  def is_visible(self) -> bool:
    return self._visible


def _make_button_menu_item(
    button_name: str,
    switchboard_inst: switchboard_base.SwitchboardBase,
    menu_container: widgets.MenuContainer) -> widgets.MenuItem:
  """Creates a menu item corresponding to a button."""
  button_menu_item = widgets.MenuItem(button_name)

  def button_handler() -> None:
    """Toggles the button and updates the button menu item text."""
    pressed_suffix = " (pressed)"
    is_pressed = button_menu_item.text.endswith(pressed_suffix)
    if is_pressed:
      switchboard_inst.release(button_name)
      button_menu_item.text = button_name
    else:
      switchboard_inst.press(button_name)
      button_menu_item.text = button_name + pressed_suffix
    # Clicking a menu item closes the menu by default and moves focus elsewhere.
    # Button press is a 2-step sequence (press & release), so don't close the
    # menu on button item clicks (by manually refocusing the menu container).
    application_current.get_app().layout.focus(menu_container.window)

  button_menu_item.handler = button_handler
  return button_menu_item


def _make_ui_window(
    window: console_config.WindowType,
    switchboard_inst: switchboard_base.SwitchboardBase
) -> Tuple[widgets.TextArea, widgets.Frame]:
  """Creates a text area and an enclosing frame.

  Args:
    window: Console window description.
    switchboard_inst: Switchboard capability instance.

  Returns:
    A tuple containing the window's text area and its enclosing UI frame.
  """
  if isinstance(window, console_config.CommandInputWindow):

    def handle_command_input(input_buffer: buffer.Buffer) -> None:
      command = input_buffer.text
      switchboard_inst.send(command, port=window.transport)

    command_input_area = widgets.TextArea(
        text="",
        multiline=False,
        accept_handler=handle_command_input,
        history=None,
        focusable=True,
        read_only=False,
        get_line_prefix=lambda line_num, wrap_count: " > "
    )
    command_input_area_bindings = key_binding.KeyBindings()

    @command_input_area_bindings.add("c-c")
    def _handle_control_c(event) -> None:  # pylint: disable=unused-variable
      del event  # Unused.
      command_input_area.text = ""
      if window.reopen_transport_on_ctrl_c:
        switchboard_inst.add_log_note(
            "The transport is running in non-interactive mode. "
            "Ctrl-C keystrokes are ignored by the device. "
            "Reopening the command transport as a workaround. "
            "Shell environment variables will be lost.")
        switchboard_inst.close_transport(port=window.transport)
        switchboard_inst.open_transport(port=window.transport)
      else:
        switchboard_inst.send(b"\x03", port=window.transport)

    command_input_area.control.key_bindings = command_input_area_bindings
    command_input_frame = widgets.Frame(
        body=command_input_area,
        title=window.name)
    return command_input_area, command_input_frame

  else:  # CommandResponseWindow or LogWindow
    search_toolbar = widgets.SearchToolbar()
    text_area = _AppendableTextArea(
        text="",
        focusable=True,
        read_only=True,
        line_numbers=True,
        scrollbar=True,
        search_field=search_toolbar,
    )
    text_area.window.ignore_content_height = lambda: True
    frame = widgets.Frame(
        body=containers.HSplit([
            text_area,
            search_toolbar,
        ]),
        title=window.name)
    return text_area, frame


class ConsoleApp:
  """Prompt_toolkit console application."""

  def __init__(
      self,
      switchboard_inst: switchboard_base.SwitchboardBase,
      device_log_file_name: str,
      configuration: console_config.ConsoleConfiguration,
      console_input: Optional[prompt_toolkit_input.Input] = None,
      console_output: Optional[output.Output] = None) -> None:
    """Initializes a prompt_toolkit console application.

    Args:
      switchboard_inst: Switchboard capability instance.
      device_log_file_name: The device log file to read (stream) logs from.
      configuration: Console configuration.
      console_input: Console input. Defaults to stdin.
      console_output: Console output. Defaults to stdout.
    """
    self._device_log_file_name = device_log_file_name
    self._window_line_transforms = configuration.window_line_transforms
    self._line_to_window_id = configuration.line_to_window_id
    self._log_file = None  # Set by run().

    body_frames = []
    self._window_id_to_text_area = collections.OrderedDict()
    for window in configuration.windows:
      text_area, enclosing_frame = _make_ui_window(window, switchboard_inst)
      self._window_id_to_text_area[window.window_id] = text_area
      body_frames.append(enclosing_frame)

    self._menu_exit_item = widgets.MenuItem("Exit", handler=_do_exit)
    if body_frames:
      self._default_focused_element = body_frames[0]
    else:
      self._default_focused_element = self._menu_exit_item
    help_window_visibility = _HelpWindowVisibility(
        focus_element_on_hide=self._default_focused_element)
    close_help_button = widgets.Button(
        text="Ok",
        handler=help_window_visibility.hide)
    help_window_visibility.focus_element_on_show = close_help_button
    help_float_dialog = containers.Float(
        containers.ConditionalContainer(
            content=widgets.Dialog(
                title="Console help",
                body=widgets.TextArea(
                    text=_HELP_TEXT,
                    focusable=True,
                    read_only=True,
                    scrollbar=True,
                    width=_HELP_TEXT_WIDTH,
                ),
                buttons=[close_help_button],
            ),
            filter=filters.Condition(help_window_visibility.is_visible),
        ),
    )

    root_container = widgets.MenuContainer(
        body=containers.FloatContainer(
            content=containers.HSplit(body_frames),
            floats=[help_float_dialog],
        ),
        menu_items=[
            self._menu_exit_item,
            widgets.MenuItem("Help", handler=help_window_visibility.show),
        ],
    )

    button_list = switchboard_inst.button_list
    button_names = button_list[0].valid_buttons() if button_list else []
    if button_names:
      button_menu_items = [
          _make_button_menu_item(button_name, switchboard_inst, root_container)
          for button_name in button_names
      ]
      self._button_dropdown = widgets.MenuItem(
          "Buttons", children=button_menu_items)
      root_container.menu_items.append(self._button_dropdown)
    else:
      self._button_dropdown = None

    # Global key bindings.
    bindings = key_binding.KeyBindings()
    bindings.add("tab")(focus.focus_next)
    bindings.add("s-tab")(focus.focus_previous)

    style = styles.Style.from_dict({
        "menu-bar": "bg:#aaaaaa #888888",
    })

    self._application = application.Application(
        layout=layout.Layout(
            container=root_container,
            focused_element=self._default_focused_element,
        ),
        style=style,
        full_screen=True,
        key_bindings=bindings,
        editing_mode=enums.EditingMode.VI,
        input=console_input,
        output=console_output,
    )

  def run(self) -> None:
    """Runs the console application until completion."""
    with open(self._device_log_file_name) as log_file:
      self._log_file = log_file
      eventloop_before = asyncio.get_event_loop()
      try:
        inputhook.set_eventloop_with_inputhook(self._input_hook)
        self._application.run()
      finally:
        asyncio.set_event_loop(eventloop_before)  # Remove the input hook.
        self._log_file = None

  def _add_text_to_window(self, text: str, window_id: int) -> None:
    """Adds text to the specified console window.

    If there's no console window with the given window_id, the text is dropped.

    Args:
      text: Text to add.
      window_id: ID of the window to add the text to.
    """
    if window_id not in self._window_id_to_text_area:
      logger.debug(f"There is no console window with ID {window_id}. "
                   f"Unable to display text: {text!r}.")
      return

    text_area = self._window_id_to_text_area[window_id]
    if isinstance(text_area, _AppendableTextArea):
      text_area.append_text(text)
    else:
      new_text = text_area.text + text
      text_area.document = document.Document(new_text, len(new_text))

  def _input_hook(self, input_hook_context: inputhook.InputHookContext) -> None:
    """Connects external I/O (device log file) to prompt_toolkit.

    Args:
      input_hook_context: Input hook context passed from prompt_toolkit. Unused.
    """
    del input_hook_context  # Unused.
    # Input hooks are supposed to yield control to the GUI event loop whenever
    # the GUI event loop has something to process. However, it seems that
    # efficiently processing all available log lines whenever they are available
    # works better than trying to yield control (which is expensive due to
    # overhead?). This implementation has to run quickly so that the GUI event
    # loop is not blocked for long. Note that this input hook is only called
    # from the GUI event loop whenever the event loop is idle.
    new_logs = self._log_file.readlines()
    self._process_device_logs(new_logs)

  def _process_device_logs(self, lines: Sequence[str]) -> None:
    """Appends device log files lines to appropriate console windows.

    Updating text windows is expensive, likely O(current_window_size). We can
    speed up the process by processing all logs in a batch and calling update
    for each window only once. This only really matters during startup, where
    thousands of log lines can be read immediately, and an inefficient
    implementation could block the UI for a few seconds.

    Args:
      lines: Text lines read from the device log file.
    """
    window_to_lines = {}
    for log_or_command_line in lines:
      window_id = self._line_to_window_id(log_or_command_line)
      if window_id not in window_to_lines:
        window_to_lines[window_id] = []
      window_to_lines[window_id].append(log_or_command_line)
    for window_id, lines in window_to_lines.items():
      transformed_lines = []
      for line in lines:
        transformed_line = line
        for before_regex, replacement in self._window_line_transforms:
          transformed_line = re.sub(before_regex, replacement, transformed_line)
        transformed_lines.append(transformed_line)
      # Lines are assumed to end in newline characters.
      self._add_text_to_window("".join(transformed_lines), window_id)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Digital Loggers Web Power Switch."""
from gazoo_device import console_config
from gazoo_device import decorators
from gazoo_device import detect_criteria
from gazoo_device import gdm_logger
from gazoo_device.base_classes import auxiliary_device
from gazoo_device.capabilities import embedded_script_dli_powerswitch
from gazoo_device.capabilities import switch_power_dli_powerswitch
from gazoo_device.switchboard import log_process
from gazoo_device.switchboard import switchboard
from gazoo_device.utility import deprecation_utils
from gazoo_device.utility import host_utils
from gazoo_device.utility import http_utils
import requests

logger = gdm_logger.get_logger()

CLASS_NAME = "DliPowerSwitch"
AUTH = requests.auth.HTTPDigestAuth("admin", "1234")
HTTP_FRAMEWORK = "http://{ip}/restapi/"
GET_CONFIG = HTTP_FRAMEWORK + "config/="

HEADERS = {
    "GET_PROP": {
        "Accept": "application/json"
    },
    "SCRIPT": {
        "Accept": "application/json",
        "Content-Type": "application/json",
        "X-CSRF": "x"
    },
    "SET_PROP": {
        "Accept": "application/json",
        "X-CSRF": "x",
        "X-HTTP-Method": "PUT"
    },
}

COMMANDS = {
    "FIRMWARE_VERSION": GET_CONFIG + "version/",
    "SERIAL_NUMBER": GET_CONFIG + "serial/",
    "DEVICE_MANUFACTURER": GET_CONFIG + "brand_company_name/",
    "DEVICE_NAME": GET_CONFIG + "brand_name/",
    "ADJUST_PORTS_MODE": HTTP_FRAMEWORK + "relay/outlets/{}/state/"
}

REGEXES = {"COMMAND_RESPONSE": ["[", '"', "]"]}


class DliPowerSwitch(auxiliary_device.AuxiliaryDevice):
  """Device class for a DLI Power Switch."""
  COMMUNICATION_TYPE = "SshComms"
  DETECT_MATCH_CRITERIA = {detect_criteria.SshQuery.is_dli: True}
  DEVICE_TYPE = "powerswitch"
  _MODEL = "LPC9"
  _NUMBER_OF_PORTS = 8
  _VALID_RETURN_CODES = [200, 206, 207]
  _OWNER_EMAIL = "gdm-authors@google.com"

  def __init__(self,
               manager,
               device_config,
               log_file_name=None,
               log_directory=None):
    super().__init__(
        manager,
        device_config,
        log_file_name=log_file_name,
        log_directory=log_directory)
    self._regexes.update(REGEXES)
    self._commands.update(COMMANDS)

  @decorators.PersistentProperty
  def ip_address(self):
    """Global IP address."""
    return self.communication_address

  @decorators.DynamicProperty
  def firmware_version(self):
    """Version of Power Switch.

    Returns:
        str: Power Switch firmware version.
    """
    return self._write_command(
        "GET", self.commands["FIRMWARE_VERSION"].format(ip=self.ip_address),
        HEADERS["GET_PROP"])

  def get_console_configuration(self) -> console_config.ConsoleConfiguration:
    """Returns the interactive console configuration."""
    return console_config.get_log_response_separate_port_configuration(
        self.switchboard.get_line_identifier())

  @decorators.CapabilityDecorator(
      switch_power_dli_powerswitch.SwitchPowerDliPowerswitch)
  def switch_power(self):
    """Returns a switch_power object to send commands.

    Returns:
        object: switch_power_dli_powerswitch
    """
    return self.lazy_init(
        switch_power_dli_powerswitch.SwitchPowerDliPowerswitch,
        http_fn=self._write_command,
        ip_address=self.ip_address,
        command_dict=self.commands,
        headers_dict=HEADERS,
        device_name=self.name,
        total_ports=self._NUMBER_OF_PORTS)

  @decorators.CapabilityDecorator(
      embedded_script_dli_powerswitch.EmbeddedScriptDliPowerswitch)
  def embedded_script(self):
    """Capability for executing embedded LUA scripts.

    Returns:
        object: embedded_script_dli_powerswitch
    """
    return self.lazy_init(
        embedded_script_dli_powerswitch.EmbeddedScriptDliPowerswitch,
        http_fn=self._write_command,
        ip_address=self.ip_address,
        base_url=HTTP_FRAMEWORK,
        device_name=self.name)

  @decorators.PersistentProperty
  def health_checks(self):
    """Returns list of methods to execute as health checks."""
    return [self.check_device_connected]

  @decorators.LogDecorator(logger)
  def get_detection_info(self):
    """Gets the persistent and optional attributes of a DLI Power Switch.

    Returns:
      tuple: (dict, dict) dictionary of persistent attributes,
              dictionary of optional attributes.

    Notes:
        persistent: model, serial_number
    """
    self.props["optional"] = {}
    self.props["persistent_identifiers"]["model"] = self._MODEL
    self.props["persistent_identifiers"]["serial_number"] = self._write_command(
        "GET",
        self.commands["SERIAL_NUMBER"].format(ip=self.ip_address),
        headers=HEADERS["GET_PROP"])
    return self.props["persistent_identifiers"], self.props["optional"]

  @classmethod
  def is_connected(cls, device_config):
    """Checks whether or not the DLI Power Switch is connected.

    Args:
      device_config (dict): contains "persistent" dict

    Returns:
        bool: whether the device responds to a ping.
    """
    return host_utils.is_pingable(
        device_config["persistent"]["console_port_name"])

  @decorators.CapabilityDecorator(switchboard.SwitchboardDefault)
  def switchboard(self):
    """Instance for communicating with the device."""
    switchboard_name = self._get_private_capability_name(
        switchboard.SwitchboardDefault)
    if not hasattr(self, switchboard_name):
      switchboard_kwargs = self._COMMUNICATION_KWARGS.copy()
      switchboard_kwargs.update({
          "communication_address": self.communication_address,
          "communication_type": self.COMMUNICATION_TYPE,
          "log_path": self.log_file_name,
          "device_name": self.name,
          "event_parser": None})
      setattr(self, switchboard_name,
              self.get_manager().create_switchboard(**switchboard_kwargs))

    return getattr(self, switchboard_name)

  def _write_command(self, method, url, headers, data=None, json_data=None):
    """Sends an HTTP request via specified method.

    Args:
        method (str): Method of HTTP request. Choices: "GET", "POST".
        url (str): HTTP formated command to send to the device.
        headers (dict): Headers required for the HTTP request.
        data (dict): Data that is needed for the HTTP POST Request
        json_data (object): JSON data that is needed for the HTTP POST Request

    Raises:
        RuntimeError: if response.status_code returned by requests.get or
        requests.post is not in valid_return_codes.
        TypeError: if headers is not a dictionary or None.

    Returns:
        str: Formatted GET/POST HTTP response
    """
    log_message = ("Sending command '{command}' to {device} via http {method} "
                   "method.").format(
                       command=url, device=self.name, method=method)
    self._log_to_file(log_message)
    try:
      if "GET" in method:
        response = http_utils.send_http_get(
            url,
            auth=AUTH,
            headers=headers,
            valid_return_codes=self._VALID_RETURN_CODES)
      else:
        response = http_utils.send_http_post(
            url,
            auth=AUTH,
            headers=headers,
            data=data,
            json_data=json_data,
            valid_return_codes=self._VALID_RETURN_CODES)

    except (RuntimeError, TypeError) as err:
      log_message = "Command '{command}' failed".format(command=url)
      self._log_to_file(log_message)
      raise err

    log_message = ("Command '{command}' sent successfully. Return Code: "
                   "{return_code}").format(
                       command=url, return_code=response.status_code)
    self._log_to_file(log_message)
    return self._format_response(response)

  @classmethod
  def _format_response(cls, response):
    """Removes extraneous formatting from requests library response.text.

    Args:
        response (requests.Response): requests library response object

    Returns:
        str: returns a formatted string of response.text
    """
    return "".join(
        i for i in response.text if i not in REGEXES["COMMAND_RESPONSE"])

  def _log_to_file(self, log_message, port="M"):
    """Adds a log message for each command."""
    if not log_message.endswith("\n"):
      log_message += "\n"
    with open(self.log_file_name, "a") as out_file:
      logline = log_process._add_log_header(log_message, port=port)
      out_file.write(logline)

  def _list_properties_dynamic_powerswitch(self):
    """Returns a set of dynamic properties.

    Returns:
        set: dynamic properties that can be queried during get-prop CLI/API
        operation.
    """
    return set(["firmware_version"])


deprecation_utils.add_deprecated_attributes(
    DliPowerSwitch,
    [("powerswitch_power", "switch_power", False),
     ("do_and_expect", "switchboard.do_and_expect", True),
     ("expect", "switchboard.expect", True),
     ("send", "switchboard.send", True),
     ("send_and_expect", "switchboard.send_and_expect", True)])



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""scp flavor of file transfer capability (for devices using SSH transports)."""
import os.path
from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.capabilities.interfaces import file_transfer_base
from gazoo_device.utility import host_utils

logger = gdm_logger.get_logger()


class FileTransferScp(file_transfer_base.FileTransferBase):
  """scp-based file or folder transfer capability."""

  def __init__(self,
               ip_address_or_fn,
               device_name,
               add_log_note_fn,
               user="root",
               key_info=None):
    """Initialize scp file or folder transfer capability.

    Args:
        ip_address_or_fn (str or callable): IP address of the device, or a
          callable which returns the IP address.
        device_name (str): name of the device using the capability.
        add_log_note_fn (func): add output to log file.
        user (str): username which scp should use.
        key_info (data_types.KeyInfo): SSH key to use.
    """
    super(FileTransferScp, self).__init__(device_name=device_name)

    self._ip_address_or_fn = ip_address_or_fn
    self._user = user
    self._key_info = key_info
    self._add_log_note_fn = add_log_note_fn

  @decorators.CapabilityLogDecorator(logger)
  def recv_file_from_device(self, src, dest="./"):
    """Copies srcs on device to dest on host.

    Args:
        src (str): file or folder path on device to copy to host.
        dest (str): destination path on host computer.

    Raises:
        DeviceError: if destination directory doesn't exist or copy failed.

    Note:
        If no dest is provided, the file will be copied to the current
        working directory on the host computer.
    """
    dest_dir = os.path.dirname(dest)
    if dest_dir != "." and not os.path.exists(dest_dir):
      raise errors.DeviceError(
          "Device {} receive from device failed. "
          "Destination directory {} doesn't appear to exist.".format(
              self._device_name, dest_dir))
    ip_address = self._get_valid_ip_address()
    logger.info("{} receiving from device. Source(s): {} Destination: {}",
                self._device_name, src, dest)
    try:
      output = host_utils.scp_from_device(
          ip_address,
          dest,
          src,
          user=self._user,
          key_info=self._key_info)
      for line in output.splitlines():
        self._add_log_note_fn(line + "\n")

      if not os.path.exists(dest):
        raise RuntimeError("Final file {} does not exist".format(dest))

    except RuntimeError as err:
      raise errors.DeviceError("Device {} receiving from device failed. "
                               "Unable to copy {} on device to {}. "
                               "Error: {!r}".format(self._device_name, src,
                                                    dest, err))

  @decorators.CapabilityLogDecorator(logger)
  def send_file_to_device(self, src, dest):
    """Copies src from host to dest on the device.

    Args:
        src (str): local file or folder path on host computer.
        dest (str): file path on device where the file should be copied to.

    Raises:
        DeviceError: if source file doesn't exist or copy failed.
    """
    if not os.path.exists(src):
      raise errors.DeviceError("Device {} send to device failed. "
                               "Source file {} doesn't appear to exist.".format(
                                   self._device_name, src))
    ip_address = self._get_valid_ip_address()
    logger.info(
        "{} sending file to device. Source: {}, destination: {}.".format(
            self._device_name, src, dest))
    try:
      output = host_utils.scp_to_device(
          ip_address,
          src,
          dest,
          user=self._user,
          key_info=self._key_info)
      for line in output.splitlines():
        self._add_log_note_fn(line + "\n")

    except RuntimeError as err:
      raise errors.DeviceError("Device {} send to device failed. "
                               "Unable to copy {} to {} on device. "
                               "Error: {!r}".format(self._device_name, src,
                                                    dest, err))

  def _check_ip_address_valid(self, ip_address):
    """Checks that the IP address of the device is valid (non-empty and pingable).

    Args:
        ip_address (str): IP address of the device.

    Raises:
        DeviceError: device has no IP address, or the IP address is not
        pingable.
    """
    if not ip_address:
      raise errors.DeviceError(
          f"{self._device_name} does not have an IPv4 address. "
          "An IPv4 address is required for file transfer functionality.")
    if not host_utils.is_pingable(ip_address):
      raise errors.DeviceError("{}'s IP address {} is not pingable.".format(
          self._device_name, ip_address))

  def _get_valid_ip_address(self):
    """Returns a valid IP address (non-empty and pingable)."""
    ip_address = self._ip_address_or_fn
    if callable(self._ip_address_or_fn):
      ip_address = self._ip_address_or_fn()
    self._check_ip_address_valid(ip_address)
    return ip_address
